// I used: http://devernay.free.fr/hacks/chip8/C8TECH10.HTM#00E0
// and https://en.wikipedia.org/wiki/CHIP-8#Opcode_table
// for information about how to implement a CHIP-8 Emulator
use chip8::emu::Chip8;
use std::path::Path;
use std::time::Duration;

use pixels::{wgpu::Surface, Error, Pixels, SurfaceTexture};
use winit::dpi::LogicalSize;
use winit::event::{Event, VirtualKeyCode};
use winit::event_loop::{ControlFlow, EventLoop};
use winit::window::WindowBuilder;
use winit_input_helper::WinitInputHelper;

static WIDTH: u32 = 64;
static HEIGHT: u32 = 32;

fn main() {
    let event_loop = EventLoop::new();
    let mut input = WinitInputHelper::new();
    let window = {
        let size = LogicalSize::new(WIDTH as f64, HEIGHT as f64);
        WindowBuilder::new()
            .with_title("Chip8 Emulator")
            .with_inner_size(size)
            .with_min_inner_size(size)
            .build(&event_loop)
            .unwrap()
    };

    let mut hidpi_factor = window.scale_factor();

    let mut pixels = {
        let surface = Surface::create(&window);
        let surface_texture = SurfaceTexture::new(WIDTH, HEIGHT, surface);
        Pixels::new(WIDTH, HEIGHT, surface_texture).unwrap()
    };

    let mut chip8: Chip8 = Default::default();

    chip8
        .load_rom(Path::new("./games/test_opcode.ch8"))
        .expect("Unable to load ROM");

    event_loop.run(move |event, _, control_flow| {
        chip8.execute_cycle();

        draw(&chip8.display.buf, pixels.get_frame());
        pixels.render().unwrap();
        window.request_redraw();

        if input.update(&event) {
            if let Some(factor) = input.scale_factor_changed() {
                hidpi_factor = factor;
            };

            if input.key_pressed(VirtualKeyCode::Escape) || input.quit() {
                *control_flow = ControlFlow::Exit;
                return;
            }

            if let Some(size) = input.window_resized() {
                pixels.resize(size.width, size.height);
            }
        }

        // std::thread::sleep(Duration::from_millis(1000));
    });
}

fn draw(chip8_gfx: &[u8], frame: &mut [u8]) {
    for (i, pixel) in frame.chunks_exact_mut(4).enumerate() {
        let rgba = if chip8_gfx[i] != 0 {
            [0xFF, 0xFF, 0xFF, 0xFF]
        } else {
            [0x00, 0x00, 0x00, 0xFF]
        };

        pixel.copy_from_slice(&rgba);
    }
}
